#define _CRT_SECURE_NO_WARNINGS

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifdef WIN32
#include <process.h>
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define lseek _lseek
#define close _close

#define O_BINARY _O_BINARY
#define ctime _ctime
#else
#include <unistd.h>
#include <stdlib.h>

#define O_BINARY 0
#endif

#include "xmalloc.h"
#include "object.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "RelationInstance.h"
#include "LandscapeClassObject.h"
#include "EntityClass.h"
#include "RelationClass.h"
#include "TAInput.h"

static CTAInput 		ta;
static CEntityInstance	*g_headP = 0;
static int				g_dead_lines = 0;

static CEntityInstance *
find_function(CEntityInstance *eP)
{
	CRelationInstance	*riP;
	CRelationClass		*rcP;
	const char			*labelP;

	for (riP = eP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		rcP    = riP->getRelationClass();
		labelP = rcP->getId();
		if (!strcmp(labelP, "contain")) {
			return(riP->m_srcP);
	}	}
	assert(0);
	return(0);
}

static void
visit(CEntityInstance *eP, int depth)
{
	CRelationInstance	*riP;
	CRelationClass		*rcP;
	const char			*labelP;
	CEntityInstance		*dstP;

	assert(!eP->m_visited);
	eP->m_visited = CalledX;
#ifdef TRACE
	printf("%p ", eP);
	for (int i = 0; i < depth; ++i) {
		fputs("  ", stdout);
	}
	printf("%s %d:%s\n", eP->getLabel(), eP->m_lineno, eP->m_fullpathP);
#endif

	for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		rcP    = riP->getRelationClass();
		labelP = rcP->getId();
		dstP   = riP->m_dstP;
		if (labelP[1] == 'E') {
			// External variable or function
			continue;
		}
		switch (labelP[0]) {
		case 'C':
			if (!strcmp(labelP, "CIF")) {
				dstP = find_function(dstP);
			} 
			break;
		} 
		if (!dstP->m_visited) {
			visit(dstP, depth+1);
}	}	}

static void
depth_search(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;

	// Find the functions and variables not visited
	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		labelP = eP->getLabel();
		if (!strcmp(labelP, "main") || !strcmp(labelP, "wmain")) {
			ecP    = eP->getEntityClass();
			labelP = ecP->getId();
			if (!labelP) {
				continue;
			}
			if (!strcmp(labelP, "GF")) {
				visit(eP, 0);
}	}	}	}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

#define ReportX	1
#define ReadX	2
#define WriteX	4
#define UpdateX	8

static int g_mode = 0;

static const char	*g_filenameP = "";
static int			g_fileno     = -1;
static char			*g_sourceP   = 0;
static int			g_source_lth = 0;

#define START_DEAD 1		// replaces \n
#define END_DEAD   2		// replaces }
#define NEWLINE    3
#define LBRACE	   4
#define RBRACE	   5
#define HASH	   6
#define END_MACRO  7		// replaces \n

static int
load_source(void)
{
	struct stat buf;
	int			ret, lth, segment;
	char		*atP, *endP;
	int			mode, state, lines, c;

	if (g_fileno >= 0) {
		close(g_fileno);
		g_fileno = -1;
	}

	if (g_mode & UpdateX) {
		mode = O_RDWR;
	} else {
		mode = O_RDONLY;
	}

	g_fileno = open(g_filenameP, mode | O_BINARY, 0);
	if (g_fileno < 0) {
		fprintf(stderr, "** Can't open %s for reading", g_filenameP);
		if (mode == O_RDWR) {
			fprintf(stderr, " and update");
		}
		fprintf(stderr, "\n");
		perror("");
		return -1;
	}
	// We have to scan the input more than once so read the entire input
	// into memory

	ret = fstat(g_fileno, &buf);
	if (ret != 0) {
		fprintf(stderr,"** fstat %s returned %d\n", g_filenameP, ret);
		perror("");
		goto bad_source;
	}

	// Allocate a large enough memory buffer

	lth = buf.st_size;
	if (!lth) {
		fprintf(stderr, "** %s is empty\n", g_filenameP);
		goto bad_source;
	}

	if (g_source_lth <= lth+2) {
		Xfree(g_sourceP);
		g_sourceP = (char *) Xmalloc(lth + 3);
		if (!g_sourceP) {
			outofmemory();
		}

		g_source_lth  = lth+2;
	}
	g_sourceP[0] = 0;
	g_sourceP[1] = '\n';
	atP          = g_sourceP + 2;
	
	// Read the source into memory

	for (endP = atP + lth; atP < endP; atP += ret) {
		segment = endP - atP;
		if (segment > SSIZE_MAX) {
			segment = SSIZE_MAX;
		}
		ret = read(g_fileno, atP, segment);
		if (ret <= 0) {
			fprintf(stderr, "** read %s returned %d\n", g_filenameP, ret);
			perror("");
bad_source:
			close(g_fileno);
			return -1;
	}	}
	assert(atP == endP);
	*endP = 0;

	if (!(g_mode & UpdateX)) {
		close(g_fileno);
	}

	state = 0;
	lines = 1;
	for (atP = g_sourceP + 2; (c = *atP); ++atP) {
		switch (c) {
		case '\n':
			++lines;
			switch (state) {
			case 1:
			case 3:
			case 4:
				*atP = NEWLINE;
				break;
			case 2:
				state = 0;
			}
			continue;
		case '\\':
			switch (state) {
			case 3:
			case 4:
				if (atP[1] && atP[1] != '\n') {
					++atP;
			}	}
			continue;
		case '\'':
			switch (state) {
			case 0:
				state = 4;
				break;
			case 4:
				state = 0;
				break;
			}
			continue;
		case '"':
			switch (state) {
			case 0:
				state = 3;
				break;
			case 3:
				state = 0;
				break;
			}
			continue;
		case '/':
			switch (state) {
			case 0:
				switch (atP[1]) {
				case '*':
					++atP;
					state = 1;
					break;
				case '/':
					++atP;
					state = 2;
					break;
				}
				break;
			}
			break;
		case '*':
			if (state == 1) {
				if (atP[1] == '/') {
					++atP;
					state = 0;
			}	}
			break;
		case '{':
			if (state) {
				*atP = LBRACE;
			}
			break;
		case '}':
			if (state) {
				*atP = RBRACE;
			}
			break;
		case '#':
			if (state) {
				*atP = HASH;
			}
			break;
	}	}
	return(0);
}

static void
add_ifdef(char *atP, int lineno) 
{
	static int show    = 0;

	char	*newlineP = 0;
	int		brackets, elses;
	char	*P;

	for (; ; --atP) {
		switch(*atP) {
		case '}':
		case '#':
		case ';':
		case 0:
		case START_DEAD:
		case END_DEAD:
		case END_MACRO:
			break;
		case '\n':
			newlineP = atP;
			continue;
		case '\\':
			// We are probably in a multi-line macro
			for (P = atP + 1; ; ++P) {
				switch(*P) {
				case ' ':
				case '\t':
				case '\r':
					continue;
				case '\n':
					while (*++P != '\n');
					newlineP = P;
					goto done_scan_backwards;
				}
				break;
			}
		default:
			continue;
		}
		break;
	}
done_scan_backwards:
	if (!newlineP) {
		fprintf(stderr, "%s:%d Can't find start of #ifdef to insert into\n", g_filenameP, lineno);
		return;
	}

	brackets = 0;
	elses    = 0;
	for (atP = newlineP; ; ++atP) {
		if (show) {
			putchar(*atP);
		}
		switch (*atP) {
		case 0:
			fprintf(stderr, "%s:%d Failed to find terminating '}'\n", g_filenameP, lineno);
			return;
		case END_DEAD:
			fprintf(stderr, "%s:%d Saw prior END_DEAD marker\n", g_filenameP, lineno);
			return;
		case '#':
			if (elses) {
				if (!strncmp(atP, "#if", 3)) {
					++elses;
					break;
				}
				if (!strncmp(atP, "#endif", 6)) {
					--elses;
				}
			} else if (!strncmp(atP, "#else", 5)) {
				elses = 1;
				break;
			}
			break;
		case '}':
			if (elses) {
				break;
			}
			if (!brackets) {
				fprintf(stderr, "%s:%d No '{' before '}'\n", g_filenameP, lineno);
				return;
			}
			if (--brackets) {
				break;
			}
			*newlineP = START_DEAD;
			*atP      = END_DEAD;
			if (show) {
				putchar('\n');
			}
			while (++newlineP <= atP) {
				switch (*newlineP) {
				case START_DEAD:
				case END_DEAD:
				case END_MACRO:
				case '\n':
					++g_dead_lines;
			}	}
			return;
		case '{':
			if (!elses) {
				++brackets;
	}	}	}
}

static void
add_macro_ifdef(char *end_lineP, const char *filenameP, int lineno)
{
	char	end_line;
	char 	*atP;
	int		brackets;

	fprintf(stderr, "%s:%d doesn't have a '{' on it -- presuming macro\n", filenameP, lineno);
	switch(end_line = *end_lineP) {
	case '\n':
		*end_lineP = START_DEAD;
		break;
	case END_MACRO:
		*end_lineP = '\n';
		break;
	default:
		fprintf(stderr, "Illegal character of %c in add_macro_ifdef\n", *end_lineP);
		exit(1);
	}

	brackets = 0;
	for (atP = end_lineP + 1; ; ++atP) {
		switch (*atP) {
		case '(':
			++brackets;
			break;
		case ')':
			--brackets;
			break;
		case '\n':
			++g_dead_lines;
			if (brackets > 0) {
				break;
			}
			*atP = END_MACRO;
			return;
		case END_MACRO:
		case 0:
			fprintf(stderr, "Can't terminate ifdef around macro in %d:%s\n", lineno, filenameP);
			*end_lineP = end_line;
			return;
}	}	}

static int
update_source(void)
{
	char		filename[4096];
	const char	*filenameP;
	FILE		*F;
	int			c;
	char		*atP;
	time_t		time_now;
	char		*dateP;

	time( &time_now );
	dateP = ctime( &time_now ); 
	dateP[24] = 0;

	if (g_mode & UpdateX) {
		F         = fdopen(g_fileno, "wb");
		filenameP = g_filenameP;
	} else {

		sprintf(filename, "%s.dead", g_filenameP);
		F = fopen(filename, "wb");
		filenameP = filename;
	}

	if (!F) {
		fprintf(stderr, "%s Can't write\n", filenameP);
		perror("");
		return -1;
	}
	rewind(F);

	if (g_sourceP[1] == START_DEAD) {
		fprintf(F, "#ifdef _DEAD_ /* %s */\n", dateP);
	}
	for (atP = g_sourceP + 2; ; ++atP) {
		switch (c = *atP) {
		case 0:
			goto done;
		case START_DEAD:
			fprintf(F, "\n#ifdef _DEAD_ /* %s */\n", dateP);
			continue;
		case END_DEAD:
			fputc('}', F);
		case END_MACRO:
			fprintf(F, "\n#endif /* _DEAD_ %s */\n", dateP);
			continue;
		case NEWLINE:
			c = '\n';
			break;
		case LBRACE:
			c = '{';
			break;
		case RBRACE:
			c = '}';
			break;
		case HASH:
			c = '#';
			break;
		}
		fputc(c, F);
	}
done:
	fclose(F);
	return(0);
}


static void
report_dead(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;
	const char			*typeP;
	int					load_failed = 1;
	char				*atP, *start_braceP, *end_lineP;
	int					at_lineno, lineno;

	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		if (eP->m_visited) {
			continue;
		}
		if (strcmp(g_filenameP, eP->m_fullpathP)) {
			if (!load_failed && (g_mode & (WriteX|UpdateX))) {
				update_source();
			}
			g_filenameP = eP->m_fullpathP;
			atP         = 0;
			load_failed = 1;
			if (g_mode & ReadX) {
				load_failed = load_source();
				if (load_failed) {
					fprintf(stderr, "%s Can't read\n", g_filenameP);
				} else {
					end_lineP    = g_sourceP + 1;
					atP          = end_lineP + 1;
					at_lineno    = 1;
					start_braceP = 0;
			}	}
			if (g_mode & ReportX) {
				printf("%s\n", g_filenameP);
		}	}
		lineno = eP->m_lineno;
		if (g_mode & ReportX) {
			ecP    = (CEntityClass *) eP->getParentClass();
			typeP  = ecP->getLabel();
			labelP = eP->getLabel();
			if (lineno == NoLineno) {
				printf("\t");
			} else {
				printf("\t%d", lineno);
			}
			printf("\t%s %s not used\n", typeP, labelP);
		}
		if (lineno != NoLineno && atP) {
			for (; at_lineno < lineno; ++atP) {
				switch (*atP) {
				case '\n':
				case NEWLINE:
					++at_lineno;
					end_lineP = atP;
					continue;
				case '{':
					start_braceP = atP;
					continue;
				case '#':
				case START_DEAD:
					end_lineP = atP;
				case END_DEAD:
				case '}':
					start_braceP = 0;
					continue;
				case 0:
					fprintf(stderr, "%s doesn't have %d lines in it\n", g_filenameP, lineno);
					exit(1);
			}	}
			if (!start_braceP) {
				for (; ; ++atP) {
					switch(*atP) {
					case '{':
						start_braceP = atP;
					case 0:
					case '\n':
						break;
					default:
						continue;
					}
					break;
			}	}
			if (start_braceP) {
				add_ifdef(start_braceP, lineno);
			} else {
				add_macro_ifdef(end_lineP, g_filenameP, lineno);
	}	}	}

	if (!load_failed && (g_mode & (WriteX | UpdateX))) {
		update_source();
	}
}

/* Reduce all references to a function (in a header file) seen in amy
 * source files to a single reference. Don't want to treat a function
 * in a header as dead code unless there really are no links to it
 * from any source file
 */

static void
consolidate(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP, **ePP, *e1P;
	const char			*labelP;
	CRelationInstance	*riP, *nextP;
	const char			*fullpathP;
	char				*P;

	// Sort by source file name and then by lineno
	for (ecP = ta.getEntityClass(0); ecP; ecP = ecP->m_nextP) {
		labelP = ecP->getId();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF") && strcmp(labelP, "LF")) {
			continue;
		}

		for (eP = ecP->m_entityInstancesP; eP; eP = eP->m_nextInClassP) {
			fullpathP = eP->m_fullpathP;
			if (!fullpathP || !strcmp(fullpathP, ".")) {
				P = 0;
				eP->buildPath(&P, 0);
				P = (char *) Xmalloc(P - ((char *) 0) + 1);
				eP->m_fullpathP = P;
				eP->buildPath(&P, 1);
			}

			for (ePP = &g_headP; (e1P = *ePP); ePP = &(e1P->m_nextInOrderP)) {
				switch (strcmp(eP->m_fullpathP, e1P->m_fullpathP)) {
				case -1:
					break;
				case 0:
					if (eP->m_lineno < e1P->m_lineno) {
						break;
					}
					if (eP->m_lineno == e1P->m_lineno) {
						// if source is compiled more than once
						// The same function might be named multiple times
						// printf("Consolidating %d:%s/%s in %s\n", eP->m_lineno, eP->getLabel(), e1P->getLabel(), eP->m_fullpathP);
						for (riP = eP->m_srcRelListP; riP; riP = nextP) {
							nextP = riP->m_next_srcP;
							// Move edge to be sourced by earlier node
							riP->m_next_srcP   = e1P->m_srcRelListP;
							e1P->m_srcRelListP = riP;
							riP->m_srcP        = e1P;
						}
						for (riP = eP->m_dstRelListP; riP; riP = nextP) {
							nextP = riP->m_next_dstP;
							// Moce edge to target earlier node
							riP->m_next_dstP   = e1P->m_dstRelListP;
							e1P->m_dstRelListP = riP;
							riP->m_dstP        = e1P;
						}
						eP->m_srcRelListP = 0;
						eP->m_dstRelListP = 0;
					
						goto skip;
					}
				default:
					continue;
				}
				break;
			}
			eP->m_nextInOrderP = *ePP;
			*ePP               = eP;
skip: 		continue;
	}	}	
}

static void
find_dead(const char *filenameP)
{
	ta.loadTA(filenameP);
	consolidate();
	depth_search();
	report_dead();
}

int 
main(int argc, const char *argv[])
{
	const char *filenameP = 0;
	int			i;
	const char	*argP;

	for (i = 1; i < argc; ++i) {
		argP = argv[i];
		switch (argP[0]) {
		case '-':
			switch (argP[1]) {
			case 'r':
				g_mode |= ReadX;
				break;
			case 'w':
				g_mode |= ReadX | WriteX;
				break;
			case 'v':
				g_mode |= ReportX;
				break;
			case 'u':
				g_mode |= ReadX | UpdateX;
				break;
			default:
				fprintf(stderr, "Unknown option %s\n", argP);
				fprintf(stderr, "usage: %s [-r/ead] [-w/rite|-u/pdate] [-v/erbose] [<] <tafile>\n", argv[0]);
				return(1);
			}
			break;
		default:
			if (filenameP) {
				fprintf(stderr, "Multiple ta_files\n");
				fprintf(stderr, "usage: %s [<] <tafile>\n", argv[0]);
				return(1);
			}
			filenameP = argP;
			break;
	}	}
	if (!g_mode) {
		g_mode = ReportX;
	}
	find_dead(filenameP);
	if (g_dead_lines) {
		printf("%d lines of source code marked dead\n", g_dead_lines);
	}
	return 0;
}

